/**
  *  \file peng.c
  *  \brief Portable Shiplist Compiler
  *
  *  This is a portable "Penguin" replacement, hacked together in a
  *  few hours.
  *
  *  (c) 2002,2003,2005,2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `README.html' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>
#include "peng.h"

#if !(defined(__STDC_VERSION__) && __STDC_VERSION__ > 199900L)
/* pre-C99 */
#  ifdef __GNUC__
#    define inline __inline__
#  else
#    define inline static
#  endif
#endif

static const char standard_footer[] =
    ">% Parsing stops here.\n"
    "Local Variables:\n"
    "font-lock-keywords: ((\">[a-z]\\\\[\\\\([^]]*\\\\)\\\\]\" 1 font-lock-string-face) "
    "(\"^%[ a-z]+\" 0 font-lock-preprocessor-face) "
    "(\">[a-z]\" 0 font-lock-keyword-face))\n"
    "End:\n";

static char nonascii = 0;
static char underscore = 2;     /* 2 means not decided yet */
static char xml_quote = 0;      /* 1 means yes */

static long playerrace[11] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };

/******************************** Truehull *******************************/

const struct Assignment truehull_defs[] = {
    { TINT,  0, 0 }, { TINT,  2, 0 }, { TINT,  4, 0 }, { TINT,  6, 0 },
    { TINT,  8, 0 }, { TINT, 10, 0 }, { TINT, 12, 0 }, { TINT, 14, 0 },
    { TINT, 16, 0 }, { TINT, 18, 0 }, { TINT, 20, 0 }, { TINT, 22, 0 },
    { TINT, 24, 0 }, { TINT, 26, 0 }, { TINT, 28, 0 }, { TINT, 30, 0 },
    { TINT, 32, 0 }, { TINT, 34, 0 }, { TINT, 36, 0 }, { TINT, 38, 0 },
};

const char truehull_footer[] =
    "\n"
    "# You must remove the hash (\"#\") for this option to take effect\n"
    "# PlayerRace = 1,2,3,4,5,6,7,8,9,10,11\n\n";
const char truehull_xml_footer[] =
    " <!-- Uncomment the following to make the PlayerRace have any effect -->\n"
    " <!--\n"
    "  <player n=\"1\" race=\"1\" />\n"
    "  <player n=\"2\" race=\"2\" />\n"
    "  <player n=\"3\" race=\"3\" />\n"
    "  <player n=\"4\" race=\"4\" />\n"
    "  <player n=\"5\" race=\"5\" />\n"
    "  <player n=\"6\" race=\"6\" />\n"
    "  <player n=\"7\" race=\"7\" />\n"
    "  <player n=\"8\" race=\"8\" />\n"
    "  <player n=\"9\" race=\"9\" />\n"
    "  <player n=\"10\" race=\"10\" />\n"
    "  <player n=\"11\" race=\"11\" />\n"
    " -->\n"
    "</truehull>\n";

const struct OutputFormat truehull_out[] = {
    { { 0, "% truehull\n\n" },
      "Race number    >a[ $a ]\n"
      "Hull slot  1   >b[ $b ]\n"
      "Hull slot  2   >c[ $c ]\n"
      "Hull slot  3   >d[ $d ]\n"
      "Hull slot  4   >e[ $e ]\n"
      "Hull slot  5   >f[ $f ]\n"
      "Hull slot  6   >g[ $g ]\n"
      "Hull slot  7   >h[ $h ]\n"
      "Hull slot  8   >i[ $i ]\n"
      "Hull slot  9   >j[ $j ]\n"
      "Hull slot 10   >k[ $k ]\n"
      "Hull slot 11   >l[ $l ]\n"
      "Hull slot 12   >m[ $m ]\n"
      "Hull slot 13   >n[ $n ]\n"
      "Hull slot 14   >o[ $o ]\n"
      "Hull slot 15   >p[ $p ]\n"
      "Hull slot 16   >q[ $q ]\n"
      "Hull slot 17   >r[ $r ]\n"
      "Hull slot 18   >s[ $s ]\n"
      "Hull slot 19   >t[ $t ]\n"
      "Hull slot 20   >u[ $u ]\n\n\n",
      { standard_footer, "\n\n" },
      "truedat.txt" },
    { { 0, 0 }, 0, { 0, 0 }, 0 },
    { { 0, "% truehull\n\n" },
      "Player  = $a\n"
      "Ships   = $b,$c,$d,$e,$f,$g,$h,$i,$j,$k,$l,$m,$n,$o,$p,$q,$r,$s,$t,$u\n",
      { truehull_footer, truehull_footer },
      "truedat.txt" },
    { { "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
        "<!DOCTYPE truehull SYSTEM \"peng.dtd\">\n"
        "<truehull>\n",
        "<truehull>\n" },
      " <race n=\"$a\">\n"
      "  <racehull n=\"$b\" />\n"
      "  <racehull n=\"$c\" />\n"
      "  <racehull n=\"$d\" />\n"
      "  <racehull n=\"$e\" />\n"
      "  <racehull n=\"$f\" />\n"
      "  <racehull n=\"$g\" />\n"
      "  <racehull n=\"$h\" />\n"
      "  <racehull n=\"$i\" />\n"
      "  <racehull n=\"$j\" />\n"
      "  <racehull n=\"$k\" />\n"
      "  <racehull n=\"$l\" />\n"
      "  <racehull n=\"$m\" />\n"
      "  <racehull n=\"$n\" />\n"
      "  <racehull n=\"$o\" />\n"
      "  <racehull n=\"$p\" />\n"
      "  <racehull n=\"$q\" />\n"
      "  <racehull n=\"$r\" />\n"
      "  <racehull n=\"$s\" />\n"
      "  <racehull n=\"$t\" />\n"
      "  <racehull n=\"$u\" />\n"
      " </race>\n",
      { truehull_xml_footer, truehull_xml_footer },
      "truehull.xml" }
};

struct FileFormat truehull_fmt = {
    40,                         /* 20 words/block */
    11,                         /* 11 races */
    0,
    20,
    truehull_defs,
    truehull_out,
    "truehull.dat",
    "truehull",
    0,
    "Player",
    0                           /* not checksummed */
};

/******************************** Engspec ********************************/

const struct Assignment engspec_defs[] = {
    { 20,     0, "Name" },         /* name */
    { TINT,  20, "Cost" },         /* cost */
    { TINT,  22, "Tritanium" },    /* t */
    { TINT,  24, "Duranium" },     /* d */
    { TINT,  26, "Molybdenum" },   /* m */
    { TINT,  28, "Techlevel" },    /* tech */
    { TLONGARRAY, 30, "FuelUsage" },    /* warp1 */
    { TLONG, 34, 0 },              /* warp2 */
    { TLONG, 38, 0 },              /* warp3 */
    { TLONG, 42, 0 },              /* warp4 */
    { TLONG, 46, 0 },              /* warp5 */
    { TLONG, 50, 0 },              /* warp6 */
    { TLONG, 54, 0 },              /* warp7 */
    { TLONG, 58, 0 },              /* warp8 */
    { TLONG, 62, 0 },              /* warp9 */
};

const char engspec_header[] = "No. Engine Name            MC  Tri  Dur  Mol Tech\n";

const struct OutputFormat engspec_out[] = {
    { { 0, "% engines\n\n" },
      "Engine number >a[ $a ]\n"
      "\n"
      "Engine name         >b[$b]\n"
      "Cost                >c[ $c ]\n"
      "Tritanium           >d[ $d ]\n"
      "Duranium            >e[ $e ]\n"
      "Molybdenum          >f[ $f ]\n"
      "Techlevel           >g[ $g ]\n"
      "Fuel usage at each warp speed\n"
      "Warp 1  >h[ $h ]\n"
      "Warp 2  >i[ $i ]\n"
      "Warp 3  >j[ $j ]\n"
      "Warp 4  >k[ $k ]\n"
      "Warp 5  >l[ $l ]\n"
      "Warp 6  >m[ $m ]\n"
      "Warp 7  >n[ $n ]\n"
      "Warp 8  >o[ $o ]\n"
      "Warp 9  >p[ $p ]\n\n\n",
      { standard_footer, "\n\n" },
      "engdat.txt" },
    /* --  -------------------- ---- ---- ---- ---- ---- */
    { { engspec_header, engspec_header },
      "$2a  $b $4c $4d $4e $4f $4g\n",
      { 0, "\n\n" },
      "engines.txt" },
    { { 0, "% engines\n\n" },
      "Engine          = $a\n"
      "Name            = $1b\n"
      "Cost            = $c\n"
      "Tritanium       = $d\n"
      "Duranium        = $e\n"
      "Molybdenum      = $f\n"
      "Techlevel       = $g\n"
      "FuelUsage       = 1 , $h\n"
      "FuelUsage       = 2 , $i\n"
      "FuelUsage       = 3 , $j\n"
      "FuelUsage       = 4 , $k\n"
      "FuelUsage       = 5 , $l\n"
      "FuelUsage       = 6 , $m\n"
      "FuelUsage       = 7 , $n\n"
      "FuelUsage       = 8 , $o\n"
      "FuelUsage       = 9 , $p\n\n",
      { 0, "\n\n" },
      "engdat.txt" },
    { { "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
        "<!DOCTYPE engines SYSTEM \"peng.dtd\">\n"
        "<engines>\n",
        "<engines>\n" },
      " <engine n=\"$a\">\n"
      "  <name>$1b</name>\n"
      "  <cost>$c</cost>\n"
      "  <tri>$d</tri>\n"
      "  <dur>$e</dur>\n"
      "  <mol>$f</mol>\n"
      "  <tech>$g</tech>\n"
      "  <fuelusage>\n"
      "   <warp n=\"1\" v=\"$h\" />\n"
      "   <warp n=\"2\" v=\"$i\" />\n"
      "   <warp n=\"3\" v=\"$j\" />\n"
      "   <warp n=\"4\" v=\"$k\" />\n"
      "   <warp n=\"5\" v=\"$l\" />\n"
      "   <warp n=\"6\" v=\"$m\" />\n"
      "   <warp n=\"7\" v=\"$n\" />\n"
      "   <warp n=\"8\" v=\"$o\" />\n"
      "   <warp n=\"9\" v=\"$p\" />\n"
      "  </fuelusage>\n"
      " </engine>\n",
      { "</engines>\n",
        "</engines>\n" },
      "engspec.xml" }
};

struct FileFormat engspec_fmt = {
    66,                         /* size */
    9,                          /* 9 engines */
    4,                          /* garbage */
    countof(engspec_defs),
    engspec_defs,
    engspec_out,
    "engspec.dat",
    "engines",
    0,
    "Engine",
    1                           /* checksums */
};

/******************************** Hullspec *******************************/

const struct Assignment hullspec_defs[] = {
    { 30,    0, "Name" },
    { TINT, 30, "Picture" },
    { TINT, 32, "Unused" },
    { TINT, 34, "Tritanium" },
    { TINT, 36, "Duranium" },
    { TINT, 38, "Molybdenum" },
    { TINT, 40, "FuelTank" },
    { TINT, 42, "Crew" },
    { TINT, 44, "Engines" },
    { TINT, 46, "Mass" },
    { TINT, 48, "Techlevel" },
    { TINT, 50, "CargoSpace" },
    { TINT, 52, "FighterBays" },
    { TINT, 54, "TorpedoTubes" },
    { TINT, 56, "BeamWeapons" },
    { TINT, 58, "Cost" }
};

const char hullspec_header[] =
    "Name                            Bm  T  F  Mass  Carg  En  Fuel  Crew  Tl  Tri  Dur  Mol    MC\n"
    "----------------------------------------------------------------------------------------------\n";

const struct OutputFormat hullspec_out[] = {
    { { 0, "% hulls\n\n" },
      "Ship number >a[ $a ]\n"
      "\n"
      "Hull name           >b[$b]\n"
      "Picture ID number   >c[ $c ]\n"
      "Tritanium           >e[ $e ]\n"
      "Duranium            >f[ $f ]\n"
      "Molybdenum          >g[ $g ]\n"
      "Fuel tank capacity  >h[ $h ]\n"
      "Crew                >i[ $i ]\n"
      "Engines             >j[ $j ]\n"
      "Unladen hull mass   >k[ $k ]\n"
      "Techlevel           >l[ $l ]\n"
      "Cargo room          >m[ $m ]\n"
      "Fighter bays        >n[ $n ]\n"
      "Torpedo tubes       >o[ $o ]\n"
      "Beam weapons        >p[ $p ]\n"
      "Cost                >q[ $q ]\n\n\n",
      { standard_footer, "\n\n" },
      "hulldat.txt"
    },
    /* ------------------------------  -- -- -- ----- ----- --- ----- ----- --- ---- ---- ---- ----- */
    { { hullspec_header, hullspec_header },
      "$b  $2p $2o $2n $5k $5m $3j $5h $5i $3l $4e $4f $4g $5q\n",
      { 0, "\n\n" },
      "hulls.txt" },
    { { 0, "% hulls\n\n" },
      "Hull            = $a\n"
      "Name            = $1b\n"
      "Picture         = $c\n"
      "Tritanium       = $e\n"
      "Duranium        = $f\n"
      "Molybdenum      = $g\n"
      "FuelTank        = $h\n"
      "Crew            = $i\n"
      "Engines         = $j\n"
      "Mass            = $k\n"
      "Techlevel       = $l\n"
      "CargoSpace      = $m\n"
      "FighterBays     = $n\n"
      "TorpedoTubes    = $o\n"
      "BeamWeapons     = $p\n"
      "Cost            = $q\n\n",
      { 0, "\n\n" },
      "hulldat.txt"
    },
    { { "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
        "<!DOCTYPE hulls SYSTEM \"peng.dtd\">\n"
        "<hulls>\n",
        "<hulls>\n" },
      " <hull n=\"$a\">\n"
      "  <name>$1b</name>\n"
      "  <pic>$c</pic>\n"
      "  <cost>$q</cost>\n"
      "  <tri>$e</tri>\n"
      "  <dur>$f</dur>\n"
      "  <mol>$g</mol>\n"
      "  <tech>$l</tech>\n"
      "  <fuel>$h</fuel>\n"
      "  <cargo>$m</cargo>\n"
      "  <crew>$i</crew>\n"
      "  <nengines>$j</nengines>\n"
      "  <mass>$k</mass>\n"
      "  <nbeams>$p</nbeams>\n"
      "  <ntubes>$o</ntubes>\n"
      "  <nbays>$n</nbays>\n"
      " </hull>\n",
      { "</hulls>\n",
        "</hulls>\n" },
      "hullspec.xml" }
};

struct FileFormat hullspec_fmt = {
    60,                         /* size */
    105,                        /* 105 hulls */
    0,
    countof(hullspec_defs),
    hullspec_defs,
    hullspec_out,
    "hullspec.dat",
    "hulls",
    0,
    "Hull",
    1                           /* checksums */
};

/******************************** Beamspec *******************************/

const struct Assignment beamspec_defs[] = {
    { 20,    0, "Name" },
    { TINT, 20, "Cost" },
    { TINT, 22, "Tritanium" },
    { TINT, 24, "Duranium" },
    { TINT, 26, "Molybdenum" },
    { TINT, 28, "Mass" },
    { TINT, 30, "Techlevel" },
    { TINT, 32, "Kill" },
    { TINT, 34, "Damage" }
};

const char beamspec_header[] = "No. Name                 Kill Destr   MC  Tri  Dur  Mol Mass Tech\n";

const struct OutputFormat beamspec_out[] = {
    { { 0, "% beams\n\n" },
      "Beam number >a[ $a ]\n"
      "\n"
      "Beam name           >b[$b]\n"
      "Cost                >c[ $c ]\n"
      "Tritanium           >d[ $d ]\n"
      "Duranium            >e[ $e ]\n"
      "Molybdenum          >f[ $f ]\n"
      "Mass                >g[ $g ]\n"
      "Techlevel           >h[ $h ]\n"
      "Kill                >i[ $i ]\n"
      "Damage              >j[ $j ]\n\n\n",
      { standard_footer, "\n\n" },
      "beamdat.txt"
    },
    /* -- --------------------  ---- ----- ---- ---- ---- ---- ---- ---- */
    { { beamspec_header, beamspec_header },
      "$2a $b  $4i $5j $4c $4d $4e $4f $4g $4h\n",
      { 0, "\n\n" },
      "beams.txt" },
    { { 0, "% beams\n\n" },
      "Beam            = $a\n"
      "Name            = $1b\n"
      "Cost            = $c\n"
      "Tritanium       = $d\n"
      "Duranium        = $e\n"
      "Molybdenum      = $f\n"
      "Mass            = $g\n"
      "Techlevel       = $h\n"
      "Kill            = $i\n"
      "Damage          = $j\n\n",
      { 0, "\n\n" },
      "beamdat.txt" },
    { { "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
        "<!DOCTYPE beams SYSTEM \"peng.dtd\">\n"
        "<beams>\n",
        "<beams>\n" },
      " <beam n=\"$a\">\n"
      "  <name>$1b</name>\n"
      "  <cost>$c</cost>\n"
      "  <tri>$d</tri>\n"
      "  <dur>$e</dur>\n"
      "  <mol>$f</mol>\n"
      "  <tech>$h</tech>\n"
      "  <mass>$g</mass>\n"
      "  <kill>$i</kill>\n"
      "  <damage>$j</damage>\n"
      " </beam>\n",
      { "</beams>\n",
        "</beams>\n" },
      "beamspec.xml" }
};

struct FileFormat beamspec_fmt = {
    36,                         /* size */
    10,                         /* 10 beams */
    0,
    countof(beamspec_defs),
    beamspec_defs,
    beamspec_out,
    "beamspec.dat",
    "beams",
    0,
    "Beam",
    1                           /* checksums */
};

/******************************** Torpspec *******************************/

const struct Assignment torpspec_defs[] = {
    { 20,    0, "Name" },
    { TINT, 20, "TorpedoCost" },
    { TINT, 22, "TubeCost" },
    { TINT, 24, "Tritanium" },
    { TINT, 26, "Duranium" },
    { TINT, 28, "Molybdenum" },
    { TINT, 30, "Mass" },
    { TINT, 32, "Techlevel" },
    { TINT, 34, "Kill" },
    { TINT, 36, "Damage" }
};

const char torpspec_header[] =
  /* -- -------------------- ---- ----- ----- ---- ---- --- --- ---- ---- */
    "                                       MC\n"
    "No. Name                Kill Destr  Tube Unit  Tri Dur Mol Mass Tech\n";

const struct OutputFormat torpspec_out[] = {
    { { 0, "% torps\n\n" },
      "Tube number >a[ $a ]\n"
      "\n"
      "Tube name           >b[$b]\n"
      "Cost of one torpedo >c[ $c ]\n"
      "Cost of tube        >d[ $d ]\n"
      "Tritanium           >e[ $e ]\n"
      "Duranium            >f[ $f ]\n"
      "Molybdenum          >g[ $g ]\n"
      "Mass                >h[ $h ]\n"
      "Techlevel           >i[ $i ]\n"
      "Kill                >j[ $j ]\n"
      "Damage              >k[ $k ]\n\n\n",
      { standard_footer, "\n\n" },
      "torpdat.txt"
    },
    { { torpspec_header, torpspec_header },
      "$2a $b $4j $5k $5d $4c $4e $3f $3g $4h $4i\n",
      { 0, "\n\n" },
      "torps.txt" },
    { { 0, "% torps\n\n" },
      "Tube            = $a\n"
      "Name            = $1b\n"
      "TorpedoCost     = $c\n"
      "TubeCost        = $d\n"
      "Tritanium       = $e\n"
      "Duranium        = $f\n"
      "Molybdenum      = $g\n"
      "Mass            = $h\n"
      "Techlevel       = $i\n"
      "Kill            = $j\n"
      "Damage          = $k\n\n",
      { 0, "\n\n" },
      "torpdat.txt" },
    { { "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
        "<!DOCTYPE beams SYSTEM \"peng.dtd\">\n"
        "<torps>\n",
        "<torps>\n" },
      " <torp n=\"$a\">\n"
      "  <name>$1b</name>\n"
      "  <cost>$d</cost>\n"
      "  <tri>$e</tri>\n"
      "  <dur>$f</dur>\n"
      "  <mol>$g</mol>\n"
      "  <tech>$i</tech>\n"
      "  <mass>$h</mass>\n"
      "  <kill>$j</kill>\n"
      "  <damage>$k</damage>\n"
      "  <torpcost>$c</torpcost>\n"
      " </torp>\n",
      { "</torps>\n",
        "</torps>\n" },
      "torpspec.xml" }
};

struct FileFormat torpspec_fmt = {
    38,                         /* size */
    10,                         /* 10 beams */
    90,
    countof(torpspec_defs),
    torpspec_defs,
    torpspec_out,
    "torpspec.dat",
    "torps",
    0,
    "Tube",
    1                           /* checksums */
};

/******************************* Everything ******************************/

struct FileFormat* const spec_files[] = {
    &truehull_fmt,
    &engspec_fmt,
    &hullspec_fmt,
    &beamspec_fmt,
    &torpspec_fmt
};

const char* single_file_headers[] = {
    "Ship List Definition (compile this with 'peng -rs1 <FILENAME')\n\n",
    "Ship List Summary\n\n",
    "Ship List Definition (compile this with 'peng -rs2 <FILENAME')\n\n",
    "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
    "<!DOCTYPE shiplist SYSTEM \"peng.dtd\">\n"
    "<!-- Ship List in XML Format -->\n\n"
    "<shiplist>\n"
};

const char* single_file_footers[] = {
    standard_footer,
    0,
    0,
    "</shiplist>\n"
};

struct FileFormat*const* files = spec_files;
unsigned             num_files = countof(spec_files);

/* These are used for checksum hacking using the "damaged ship"
   picture ids */
#define HULLSPEC_INDEX 2
#define HULLSPEC_FREE_POS 32
#define HULLSPEC_FREE_LEN 2

/* These are for the special values in v2 */
#define ENGSPEC_FUEL 30

/******************************* Utilities *******************************/

const char* progname;
const char* directory;
char verbose;                   /* verboseness level */

inline int
get_expected_file_size(const struct FileFormat* f)
{
    return f->block_size * f->entry_count + f->garbage_count;
}

inline int
get_int(const uchar* p)
{
    return p[0] + 256U*p[1];
}

inline long
get_long(const uchar* p)
{
    return (unsigned)get_int(p) + 65536UL*(unsigned)get_int(p+2);
}

inline void
store_int(uchar* p, unsigned int i)
{
    *p++ = i & 255;
    *p   = (i >> 8) & 255;
}

inline void
store_long(uchar* p, unsigned long i)
{
    store_int(p, i & 65535);
    store_int(p+2, (i >> 16) & 65535);
}

int
get_array_size(const struct FileFormat* fmt, const struct Assignment* asgn)
{
    int i = asgn - fmt->definitions, i0 = i;
    do
        ++i;
    while (i < fmt->field_count && !fmt->definitions[i].name);
    return i - i0;
}

void
die(const char* f, ...)
{
    va_list l;
    va_start(l, f);
    fprintf(stderr, "%s: ", progname);
    vfprintf(stderr, f, l);
    fprintf(stderr, "\n");
    exit(EXIT_FAILURE);
}

void
log(int level, const char* f, ...)
{
    va_list l;
    va_start(l, f);
    if (verbose >= level) {
        fprintf(stdout, "%*s", 4*(level-1), "");
        vfprintf(stdout, f, l);
        fprintf(stdout, "\n");
    }
}

void
help(void)
{
    printf("%s %s -- Portable Ship List Compiler/Decompiler for VGA Planets(tm)\n"
           "\n"
           "usage: %s [-drtVh] [-vvvfa] [-cx] [-12] [-uU] [-s] [directory]\n"
           "\n"
           "-d     decompile ship list (.dat -> .txt)\n"
           "-r     recompile ship list (.txt -> .dat)\n"
           "-t     make nice tables from ship list (.dat -> .txt)\n"
           "-X     make XML files from ship list (.dat -> .xml)\n"
           "-1/-2  select Penguin 1.x/2.x mode (default: -%d)\n"
           "-s     select single-file mode (data on stdin / stdout)\n"
           "\n"
           "-v     verbose\n"
           "-V     show version\n"
           "-h     show friendly help text\n"
           "-f     from scratch (don't load existing files)\n"
           "-c     ignore checksums\n"
           "-x     try harder with checksums\n"
           "-a     permit non-ASCII characters in names\n"
           "-u     translate space in specs to underline in text (default for -2)\n"
           "-U     do not translate space to underline (default for -1)\n"
           "\n"
           "Hacked together in Summer 2002, updated in 2005, 2007\n"
           "by Stefan Reuther <Streu@gmx.de>\n",
           progname, VERSION, progname, DEFAULT_VERSION);
    exit(0);
}

void
version(void)
{
    printf("Portable Ship List Compiler/Decompiler Version " VERSION "\n");
    exit(0);
}

/** Open a file and return fp. die() on error (never returns NULL). */
FILE*
open_file(const char* basename, const char* mode)
{
    FILE* fp;
    char* total = malloc(strlen(basename) + strlen(directory) + 2);
    if (!total)
        die("out of memory");
    sprintf(total, "%s/%s", directory, basename);
    fp = fopen(total, mode);
    if (!fp)
        die("unable to open %s: %s", total, strerror(errno));
    free(total);
    return fp;
}

/** True iff a equals b, ignoring case. */
int
str_case_eq(const char* a, const char* b)
{
    while (*a) {
        if (toupper(*a) != toupper(*b))
            return 0;
        ++a, ++b;
    }
    if (*b)
        return 0;
    else
        return 1;
}

/******************************** Loading ********************************/

void
allocate_memory(void)
{
    unsigned i;
    log(1, "Allocating memory...");
    for (i = 0; i < num_files; ++i) {
        files[i]->data = calloc(get_expected_file_size(files[i]), 1);
        if (!files[i]->data)
            die("out of memory");
    }
}

void
load_files(void)
{
    unsigned i;
    FILE* fp;
    log(1, "Loading files...");
    for (i = 0; i < num_files; ++i) {
        log(2, "%s", files[i]->dat_name);
        fp = open_file(files[i]->dat_name, "rb");
        fread(files[i]->data, 1, get_expected_file_size(files[i]), fp);
        fclose(fp);
    }
}

/***************************** Writing files *****************************/

void
save_files(void)
{
    unsigned i;
    FILE* fp;
    log(1, "Saving files...");
    for (i = 0; i < num_files; ++i) {
        unsigned size = get_expected_file_size(files[i]);
        log(2, "%s", files[i]->dat_name);
        fp = open_file(files[i]->dat_name, "wb");
        if (size != fwrite(files[i]->data, 1, size, fp))
            fprintf(stderr, "%s: write error\n", files[i]->dat_name);
        fclose(fp);
    }
}

/** Write one output record.
    \param fp       file
    \param fmt      format string with `$x' placeholders
    \param data     respective data record
    \param defs     data definitions
    \param index    record number (for $a placeholder) */
void
write_output(FILE* fp, const char* fmt, const uchar* data,
             const struct Assignment* defs, int index)
{
    const char* p;
    int wi;
    while ((p = strchr(fmt, '$'))) {
        fwrite(fmt, 1, p - fmt, fp);
        wi = 0;
        while (p[1] >= '0' && p[1] <= '9')
            wi = 10*wi + *++p-'0';
        if (p[1] == 'a') {
            fprintf(fp, "%*d", wi, index);
        } else {
            unsigned x = p[1] - 'b';
            if (defs[x].type == TINT || defs[x].type == TINTARRAY) {
                fprintf(fp, "%*d", wi, get_int(data + defs[x].offset));
            } else if (defs[x].type == TLONG || defs[x].type == TLONGARRAY) {
                fprintf(fp, "%*ld", wi, get_long(data + defs[x].offset));
            } else {
                const uchar* src = data + defs[x].offset;
                int n = defs[x].type;
                if (wi && !underscore) {
                    while (n && (src[n-1] == ' ' || src[n-1] == 0))
                        --n;
                }
                while (n--) {
                    char c = *src ? *src++ : ' ';
                    if (xml_quote && (c == '<' || c == '>' || c == '"' || c == '&')) {
                        fprintf(fp, "&#%d;", c);
                    } else {
                        if (underscore && c == ' ')
                            c = '_';
                        fputc(c, fp);
                    }
                }
            }
        }
        fmt = p + 2;
    }
    fprintf(fp, "%s", fmt);
}

void
save_output(FILE* fp, const struct OutputFormat* fmt, int file_index, int format_index)
{
    int j;
    if (fmt->title[format_index])
        fprintf(fp, "%s", fmt->title[format_index]);
    for (j = 0; j < files[file_index]->entry_count; ++j) {
        write_output(fp, fmt->format,
                     &files[file_index]->data[j * files[file_index]->block_size],
                     files[file_index]->definitions, j+1);
    }
    if (fmt->footer[format_index])
        fprintf(fp, "%s", fmt->footer[format_index]);
}

/** Create text output in format /format_idx/ */
void
save_text(int format_idx, int single_file)
{
    unsigned i;
    if (single_file) {
        /* Single-file output */
        log(1, "Writing output...");
        if (single_file_headers[format_idx])
            fprintf(stdout, "%s", single_file_headers[format_idx]);
        for (i = 0; i < num_files; ++i) {
            const struct OutputFormat* fmt = files[i]->output + format_idx;
            if (!fmt->format)
                continue;
            save_output(stdout, fmt, i, 1);
        }
        if (single_file_footers[format_idx])
            fprintf(stdout, "%s", single_file_footers[format_idx]);
    } else {
        /* Multiple-file output */
        FILE* fp;
        log(1, "Writing text files...");
        for (i = 0; i < num_files; ++i) {
            const struct OutputFormat* fmt = files[i]->output + format_idx;
            if (!fmt->txt_name)
                continue;

            log(2, "%s", fmt->txt_name);
            fp = open_file(fmt->txt_name, "w");
            save_output(fp, fmt, i, 0);
            fclose(fp);
        }
    }
}

/*************************** Reading Text Files **************************/

const struct FileFormat*
get_format_by_name(const char* name)
{
    unsigned i;
    for (i = 0; i < countof(spec_files); ++i) {
        if (strcmp(spec_files[i]->sec_name, name) == 0) {
            log(2, "Changed to section '%s'", name);
            return spec_files[i];
        }
    }
    log(2, "Changed to unknown section '%s'", name);
    return 0;
}

void
remove_underscores(char* p)
{
    while ((p = strchr(p, '_')))
        *p++ = ' ';
}

/** Prepare name on LHS of a v2 assignment (remove whitespace). */
void
prepare_name(char* p)
{
    char* src = p;
    while (*src)
        if (!isspace(*src))
            *p++ = *src++;
        else
            ++src;
    *p = 0;
}

/** Parse an integer. Rejects all syntax errors. */
long
parse_int(const char* buffer, long min, long max, const char* fn, int line)
{
    char* end;
    long value = strtol(buffer, &end, 10);
    while (*end && isspace(*end))
        ++end;
    if (!*buffer || *end)
        die("Invalid number in file %s, line %d", fn, line);
    if (value < min || value > max)
        die("Value out of range in file %s, line %d", fn, line);
    return value;
}

/** String assignment.
    \param buffer     string to assign
    \param pos        record that will receive assignment
    \param a          corresponding field definition
    \param fn,line    for error messages */
void
assign_string(char* buffer, uchar* pos, const struct Assignment* a,
              const char* fn, int line)
{
    int i;
    char* p;

    if (underscore)
        remove_underscores(buffer);

    if (!nonascii)
        for (p = buffer; *p; ++p)
            if ((unsigned char)*p > 127)
                die("Non-ASCII character not permitted in file %s, line %d (use `-a' to relax this restriction)", fn, line);

    for (i = strlen(buffer); i > 0; --i)
        if (buffer[i-1] == ' ')
            buffer[i-1] = 0;
        else
            break;
    if ((int) strlen(buffer) > a->type)
        die("String too long for this field in file %s, line %d", fn, line);
    p = buffer;
    for (i = 0; i < a->type; ++i)
        pos[a->offset + i] = *p ? *p++ : ' ';
}

/** Parse line of numeric assignments in v2 format.
    \param p       text
    \param min,max ranges
    \param fn,line for error messages
    \param values  results go here
    \param max_values maximum number of values allowed
    \returns number of values read (never zero, <= max_values) */
int
parse_v2_line(const char* p, long min, long max, const char* fn, int line,
              long* values, int max_values)
{
    int num_values = 0;
    long v;
    while (1) {
        while (*p && isspace(*p))
            ++p;
        if (!*p || *p == '#')
            break;
        if (*p < '0' || *p > '9')
            die("Invalid number in file %s, line %d", fn, line);
        if (num_values >= max_values)
            die("Too many values in file %s, line %d", fn, line);

        v = strtol(p, (char**)&p, 10);
        if (v < min || v > max)
            die("Value out of range in file %s, line %d", fn, line);
        values[num_values++] = v;

        while (*p && isspace(*p))
            ++p;
        if (*p == ',')          /* not really good. oh well. */
            ++p;
    }

    if (!num_values)
        die("No values in file %s, line %d", fn, line);

    return num_values;
}

/** Read a file in v2 format. */
void
read_v2_file(FILE* fp, const char* fn, const struct FileFormat* fmt)
{
    char  line_buffer[1024];
    char *p, *eq, *q;
    int   line = 0, n, too_long;

    int  index_lo = 0, index_hi = 0, index;
#define INDEX_LOOP(XX) for(index = index_lo, XX = fmt->data + (index_lo-1) * fmt->block_size; \
                           index <= index_hi; \
                           ++index, XX += fmt->block_size)

    while (fgets(line_buffer, sizeof line_buffer, fp)) {
        /* I could have made an `agets'... */
        ++line;
        n = strlen(line_buffer);
        too_long = 0;
        if (n && line_buffer[n-1] != '\n') {
            /* no \n in buffer. Either user didn't enter one, or
               line is too long for buffer. */
            too_long = n > (int)(sizeof(line_buffer)-2);
            while ((n = fgetc(fp)) != EOF && n != '\n')
                ;
        } else {
            /* remove trailing \r, \032 and whatever might be there */
            while (n && (unsigned char)line_buffer[n-1] < ' ')
                line_buffer[--n] = 0;
        }

        p = line_buffer;
        while (*p && isspace(*p))
            ++p;
        if (!*p || *p == '#')
            continue;
        if (too_long)
            die("Line %d is too long in file %s", line, fn);
        if (*p == '%') {
            /* section change */
            ++p;
            while (*p && isspace((unsigned char) *p))
                ++p;
            q = p;
            while (isalnum((unsigned char) *q))
                ++q;
            if (p != q) {
                *q = 0;
                for (index = 0; p[index] != 0; ++index)
                    p[index] = tolower((unsigned char) p[index]);
                fmt = get_format_by_name(p);
            } else {
                die("Invalid section change in file %s, line %d", fn, line);
            }
            continue;
        }
        if (!fmt)
            continue;

        /* parse it */
        eq = strchr(p, '=');
        if (!eq)
            die("No `=' in file %s, line %d", fn, line);
        *eq++ = 0;
        while (*eq && isspace(*eq))
            ++eq;

        /* now, eq points to the value and p points to the name,
           with optional spaces */
        prepare_name(p);

        if (str_case_eq(p, fmt->item_name)) {
            /* we're assigning the current position */
            long values[2];
            int res = parse_v2_line(eq, 1, fmt->entry_count, fn, line, values, 2);
            if (res == 1) {
                index_lo = index_hi = values[0];
            } else if (values[0] <= values[1]) {
                index_lo = values[0];
                index_hi = values[1];
            } else {
                die("Current position assignment must be in order low,high in file %s, line %d", fn, line);
            }
            if (index_lo != index_hi)
                log(3, "Pointer set to %d..%d", index_lo, index_hi);
            else
                log(3, "Pointer set to %d", index_hi);
            continue;
        } else if (fmt == &truehull_fmt && str_case_eq(p, "PlayerRace")) {
            /* playerrace assignment in truehull */
            int res = parse_v2_line(eq, 1, 11, fn, line, playerrace, 11);
            while (res < 11)
                playerrace[res] = playerrace[res-1], ++res;
            continue;
        }

        /* everything else is a value assigmnent, so we must have
           selected an entry */
        if (index_lo == 0)
            die("Current position not set (`%s = NN') in file %s, line %d",
                fmt->item_name, fn, line);

        if (fmt == &truehull_fmt && str_case_eq(p, "Ships")) {
            /* this one is special */
            long values[20];
            int  res = parse_v2_line(eq, 0, hullspec_fmt.entry_count, fn, line, values, 20);
            int  i;
            uchar* data;
            INDEX_LOOP(data) {
                for (i = 0; i < 20; ++i)
                    store_int(data + 2*i, i < res ? values[i] : 0);
            }
        } else {
            /* "regular" format */
            const struct Assignment* a = 0;
            int i;
            for (i = 0; i < fmt->field_count; ++i)
                if (fmt->definitions[i].name && str_case_eq(p, fmt->definitions[i].name)) {
                    a = &fmt->definitions[i];
                    break;
                }
            if (!a)
                die("Invalid name in file %s, line %d", fn, line);

            if (a->type == TINTARRAY) {
                long values[2];
                uchar* data;
                if (parse_v2_line(eq, 0, 0x7FFF, fn, line, values, 2) != 2)
                    die("Need two values for `%s' in file %s, line %d", p, fn, line);
                if (values[0] < 1 || values[0] > get_array_size(fmt, a))
                    die("First value must be between 1 and %d in file %s, line %d",
                        get_array_size(fmt, a), fn, line);
                INDEX_LOOP(data) {
                    store_int(data + 2*(values[0]-1) + a->offset, values[1]);
                }
            } else if (a->type == TLONGARRAY) {
                /* array of long, i.e. 'FuelUsage = 1, 99' */
                long values[2];
                uchar* data;
                if (parse_v2_line(eq, 0, 0x7FFFFFFF, fn, line, values, 2) != 2)
                    die("Need two values for `%s' in file %s, line %d", p, fn, line);
                if (values[0] < 1 || values[0] > get_array_size(fmt, a))
                    die("First value must be between 1 and %d in file %s, line %d",
                        get_array_size(fmt, a), fn, line);
                INDEX_LOOP(data) {
                    store_long(data + 4*(values[0]-1) + a->offset, values[1]);
                }
            } else if (a->type == TINT) {
                /* integer */
                long value;
                uchar* data;
                parse_v2_line(eq, -32767, 32767, fn, line, &value, 1);
                INDEX_LOOP(data) {
                    store_int(data + a->offset, value);
                }
            } else if (a->type == TLONG) {
                /* long */
                long value;
                uchar* data;
                parse_v2_line(eq, LONG_MIN, LONG_MAX, fn, line, &value, 1);
                INDEX_LOOP(data) {
                    store_long(data + a->offset, value);
                }
            } else {
                /* string */
                uchar* data;
                INDEX_LOOP(data) {
                    assign_string(eq, data, a, fn, line);
                }
            }
        }
    }
}

/** Read file in v1 format */
void
read_single_file(FILE* fp, const char* fn, const struct FileFormat* fmt)
{
    uchar* pos = 0;
    char assign = -1;
    char buffer[100], *p;
    int line = 1;
    int c;
    while ((c = fgetc(fp)) != EOF) {
        switch (c) {
         case '\n':
            ++line;
            break;
         case '>':
            /* set read pointer */
            assign = fgetc(fp) - 'a';
            if (assign == '%'-'a')
                return;
            if (fmt)
                if (assign < 0 || assign > fmt->field_count)
                    die("Invalid assignment in file %s, line %d", fn, line);
            break;
         case '%':
            /* section change */
            while ((c = fgetc(fp)) != EOF && c != '\n' && isspace((unsigned char) c))
                ;
            if (c == EOF || c == '\n')
                die("Expected section name after '%', line %d", line);
            p = buffer;
            *p++ = tolower(c);
            while ((c = fgetc(fp)) != EOF && c != '\n' && !isspace((unsigned char) c))
                if (p != buffer+sizeof(buffer)-1)
                    *p++ = tolower(c);
            *p = 0;
            fmt = get_format_by_name(buffer);
            assign = -1;
            break;
         case '[':
            /* assignment */
            if (fmt) {
                if (assign < 0)
                    die("Stray `[' in file %s, line %d", fn, line);
                p = buffer;
                while ((c = fgetc(fp)) != ']') {
                    if (c == EOF || c == '\n')
                        die("Unterminated `[' in file %s, line %d", fn, line);
                    if (c != '\r') {
                        if (c <= 128 && !isprint(c)) {
                            die("Control character in file %s, line %d", fn, line);
                        } else {
                            if (p >= buffer + sizeof(buffer)-1)
                                die("Value too long in file %s, line %d", fn, line);
                            *p++ = c;
                        }
                    }
                }
                *p = 0;
                /* We have filled the buffer. Attempt assignment */
                if (assign == 0) {
                    /* a brace -> set pointer */
                    int i = parse_int(buffer, 1, fmt->entry_count, fn, line);
                    pos = fmt->data + (i-1) * fmt->block_size;
                    log(3, "Pointer set to item %d", i);
                } else {
                    const struct Assignment* a = &fmt->definitions[assign-1];
                    if (!pos)
                        die("Current position (`\"a\" brace') not set in file %s, line %d", fn, line);
                    if (a->type == TINT || a->type == TINTARRAY) {
                        store_int(pos + a->offset, parse_int(buffer, -32768, +32767, fn, line));
                    } else if (a->type == TLONG || a->type == TLONGARRAY) {
                        store_long(pos + a->offset, parse_int(buffer, LONG_MIN, LONG_MAX, fn, line));
                    } else {
                        /* string */
                        assign_string(buffer, pos, a, fn, line);
                    }
                }
                assign = -1;
            }
            break;
        }
    }
}

void
read_text(int index, void (*func)(FILE*,const char*,const struct FileFormat*), int single_file)
{
    if (single_file) {
        log(1, "Reading input...");
        func(stdin, "(standard input)", 0);
    } else {
        unsigned i;
        FILE* fp;
        log(1, "Reading text files...");
        for (i = 0; i < num_files; ++i) {
            const char* name = files[i]->output[index].txt_name;
            log(2, "%s", name);
            fp = open_file(name, "r");
            func(fp, name, files[i]);
            fclose(fp);
        }
    }
}

/******************************* Checksums *******************************/

long
compute_checksum(const struct FileFormat* f)
{
    int size = get_expected_file_size(f);
    int i;
    long sum = 0;
    for (i = 0; i < size; ++i)
        sum += f->data[i];
    return sum;
}

/** Attempt to correct checksum by hacking all bytes in specified block.
    \param p,size data block
    \param dif    checksum difference (actual checksum minus desired checksum)
    \returns remaining difference */
long
hack_checksum_block(uchar* p, size_t size, long dif)
{
    size_t n = 0;
    while (dif && n < size) {
        if (dif > 0) {
            if (p[n] > dif) {
                p[n] -= dif; dif = 0;
            } else {
                dif -= p[n]; p[n] = 0;
            }
        } else {
            /* we have to increase value */
            uchar limit = 255 - p[n];
            if (limit > -dif) {
                p[n] -= dif; dif = 0;
            } else {
                dif += limit; p[n] = 255;
            }
        }
        ++n;
    }
    return dif;
}

/** Hack checksum in garbage trailer of file f. */
long
hack_checksum(const struct FileFormat* f, long dif)
{
    /* positive difference means checksum is too big and we have to
       decrease values */
    uchar* p = f->data + f->block_size * f->entry_count;
    if (!dif)
        return dif;
    dif = hack_checksum_block(p, f->garbage_count, dif);
    log(2, "%s [trailer]: remaining difference is %ld", f->dat_name, dif);
    return dif;
}

/** Hack checksums in one name.
    \param p,len  a name block
    \param dif    desired difference */
long
hack_checksum_name(uchar* p, int len, long dif)
{
    /* replace trailing spaces in one name by nulls. Start with the
       first trailing space. I.e. when the name is "Foo     ", replace
       it by "Foo\0\0\0  ". This will work nicely with both C and BASIC
       programs (the former will stop parsing at the first null, the
       latter will strip as much as possible). Because planets.exe has
       no glyph for 0, this can rarely be seen anyway. */
    int n = len;
    while (n > 0 && p[n-1] == ' ')
        --n;
    while (dif > 0 && n < len) {
        p[n] = 0;
        ++n;
        dif -= ' ';
    }
    return dif;
}

/** Hack checksums in all names */
long
hack_checksum_names(const struct FileFormat* f, long dif)
{
    /* in names, replace trailing spaces by nulls. */
    int n;
    int ofs = 0, len = -1;

    if (dif <= 0)
        return dif;

    for (n = 0; n < f->field_count; ++n) {
        if (f->definitions[n].type > 0) {
            ofs = f->definitions[n].offset;
            len = f->definitions[n].type;
            break;
        }
    }
    if (len < 0)
        /* record doesn't contain a "name" field */
        return dif;

    for (n = 0; n < f->entry_count; ++n) {
        dif = hack_checksum_name(f->data + n * f->block_size + ofs, len, dif);
        if (dif < 0)
            break;
    }
    log(2, "%s [names]: remaining difference is %ld", f->dat_name, dif);
    return dif;
}

/** In all records, hack a specified field. */
long
hack_checksum_record(const struct FileFormat* f, long dif, int ofs, int len)
{
    int i;
    if (!dif)
        return dif;
    for (i = 0; i < f->entry_count; ++i) {
        uchar* p = f->data + i * f->block_size;
        dif = hack_checksum_block(p + ofs, len, dif);
        if (!dif)
            break;
    }
    return dif;
}

void
update_checksums(int try_harder)
{
    unsigned i;
    long sum = 0;
    log(1, "Updating checksums...");
    for (i = 0; i < num_files; ++i)
        if (files[i]->checksummed)
            sum += compute_checksum(files[i]);

    sum -= MAGIC_CHECKSUM;
    log(2, "Checksum difference is %ld", sum);
    for (i = 0; i < num_files; ++i)
        if (files[i]->checksummed && files[i]->garbage_count)
            sum = hack_checksum(files[i], sum);

    if (sum && try_harder) {
        log(2, "Remaining difference is %ld, trying plan B (unused hullspec field)", sum);
        sum = hack_checksum_record(files[HULLSPEC_INDEX], sum,
                                   HULLSPEC_FREE_POS, HULLSPEC_FREE_LEN);
        if (sum > 0) {
            log(2, "Remaining difference is %ld, trying plan C (names)", sum);
            for (i = 0; i < num_files; ++i)
                if (files[i]->checksummed)
                    sum = hack_checksum_names(files[i], sum);
            for (i = 0; i < num_files; ++i)
                if (files[i]->checksummed && files[i]->garbage_count)
                    sum = hack_checksum(files[i], sum);
        }
    }

    if (sum) {
        die("Unable to fix checksums. Remaining difference is %ld.\nChanges will not be saved. See manual for help.", sum);
    }

    /* validate */
    sum = 0;
    for (i = 0; i < num_files; ++i)
        if (files[i]->checksummed)
            sum += compute_checksum(files[i]);
    if (sum != MAGIC_CHECKSUM)
        die("Program bug: checksum mismatch");
}

/*************************** PlayerRace support **************************/

void
shuffle_truehull(uchar* data)
{
    uchar data_tmp[440];
    int i;

    memcpy(data_tmp, data, 440);
    for (i = 0; i < 11; ++i)
        memcpy(data + i*40, data_tmp + (playerrace[i]-1)*40, 40);
}

/********************************** Main *********************************/

int mode = 0;
int fromscratch = 0;
int csummode = 0;
int fmt_version = DEFAULT_VERSION;
int single_file_mode = 0;

void
handle_option_letter(char c)
{
    switch(c) {
     case '1':
     case '2':
        fmt_version = c-'0';
        break;
     case 'd':
     case 'r':
     case 't':
     case 'X':
        mode = c;
        break;
     case 's':
        single_file_mode = 1;
        break;
     case 'h':
        help();
        break;
     case 'v':
        ++verbose;
        break;
     case 'c':
     case 'x':
        csummode = c;
        break;
     case 'V':
        version();
        break;
     case 'f':
        fromscratch = 1;
        break;
     case 'a':
        nonascii = 1;
        break;
     case 'u':
        underscore = 1;
        break;
     case 'U':
        underscore = 0;
        break;
     default:
        die("invalid option `%c'. Try `%s -h' for help.", c, progname);
    }
}

struct OptionMap {
    const char* name;
    char        letter;
};
static const struct OptionMap long_options[] = {
    { "decompile", 'd' },
    { "recompile", 'r' },
    { "compile", 'r' },
    { "table", 't' },
    { "xml", 'X' },
    { "single-file", 's' },
    { "help", 'h' },
    { "verbose", 'v' },
    { "ignore-checksums", 'c' },
    { "extra-checksums", 'x' },
    { "version", 'V' },
    { "from-scratch", 'f' },
    { "allow-non-ascii", 'a' },
    { "use-underscore", 'u' },
    { "no-underscore", 'U' }
};

int
main(int argc, char** argv)
{
    char* p;
    int doing_options = 1;
    (void) argc;

    /* poor man's portable "basename" */
    progname = *argv;
    while ((p = strpbrk(progname, "\\:/")))
        progname = p+1;

    while (*++argv) {
        if (doing_options && **argv == '-') {
            char* p = *argv;
            if (p[1] == '-') {
                /* long option */
                if (p[2] == 0) {
                    doing_options = 0;
                } else {
                    size_t i;
                    for (i = 0; i < countof(long_options); ++i)
                        if (strcmp(p+2, long_options[i].name) == 0)
                            break;
                    if (i < countof(long_options))
                        handle_option_letter(long_options[i].letter);
                    else
                        die("invalid option `%s'. Try `%s --help' for help.", p, progname);
                }
            } else {
                while (*++p)
                    handle_option_letter(*p);
            }
        } else if (!directory) {
            directory = *argv;
        } else {
            die("too many directory names on command line. Try `%s -h' for help.", progname);
        }
    }
    if (!mode)
        die("mode not specified (-r/-d/-t/-X). Try `%s -h' for help.", progname);
    if (!directory)
        directory = ".";
    if (underscore == 2)
        underscore = (fmt_version == 2 && mode != 't' && mode != 'X');

    /* action! */
    allocate_memory();
    if (!fromscratch)
        load_files();

    if (mode == 'd') {
        /* decompile mode */
        save_text(fmt_version == 1 ? 0 : 2, single_file_mode);
    } else if (mode == 't') {
        /* table mode */
        save_text(1, single_file_mode);
    } else if (mode == 'X') {
        /* XML mode */
        xml_quote = 1;
        save_text(3, single_file_mode);
    } else {
        /* recompile mode */
        if (fmt_version == 1) {
            read_text(0, read_single_file, single_file_mode);
        } else {
            read_text(2, read_v2_file, single_file_mode);
            shuffle_truehull(truehull_fmt.data);
        }
        if (csummode != 'c')
            update_checksums(csummode == 'x');
        save_files();
    }
    return 0;
}
